(* Finds battleships, which are rectangular sections
   of a grid where some property holds.
   
   Unlike the real battleship game, battleships of
   the same property are not allowed to be adjacent
   to one another. This means that we can do it
   pretty quickly and that there is no ambiguity.
   There are also illegal configurations, which this
   reports. *)

signature BATTLESHIP =
sig

  datatype battleship =
    (* If we see that the battleship is not rectangular,
       we return some rectangle that contains the input
       point, at least one unset point, and which cannot
       be reduced to a legal battleship (by removing an
       entire unset row or column). *)
    IllegalConfig of { x : int, y : int, w : int, h : int }
  | Battleship of { x : int, y : int, w : int, h : int }

  (* battleship (x, y, test)
     Finds a battleship. Give a function 'test' that returns
     true if a position is set. If it's reading from arrays,
     it should return false when outside the bounds. It should
     always return false near numeric limits to avoid overflow.
     The test(x, y) must be true. It will be included in the
     resultant battleship, but could easily be in its interior. 
     
     This function is linear in the size of the battleship
     (which is polynomial in its shortest edge), because it has
     to check every cell in its interior and border. *)

  val battleship : int * int * (int * int -> bool) -> battleship

end