
structure Battleship :> BATTLESHIP =
struct

  datatype battleship =
    IllegalConfig of { x : int, y : int, w : int, h : int }
  | Battleship of { x : int, y : int, w : int, h : int }

  fun util_for lo hi f =
      if lo > hi then ()
      else (ignore (f lo); util_for (lo + 1) hi f)

  exception Return of battleship
  fun battleship (startx, starty, test) =
     let
         (* First thing is to find the top-left coordinate. Keep moving
            up and left until test returns false. *)

         fun moveup (curx, cury) =
             if test (curx, cury - 1)
             then moveup (curx, cury - 1)
             else (curx, cury)

         fun moveupleft (curx, cury) =
             if test (curx - 1, cury)
             then moveupleft (curx - 1, cury)
             else moveup (curx, cury)

         val (x, y) = moveupleft (startx, starty)

         (* Now the same, to find the bottom right. *)

         fun movedown (curx, cury) =
             if test (curx, cury + 1)
             then movedown (curx, cury + 1)
             else (curx, cury)

         fun movedownright (curx, cury) =
             if test (curx + 1, cury)
             then movedownright (curx + 1, cury)
             else movedown (curx, cury)

         val (xx, yy) = movedownright (startx, starty)

         (* This is the largest possible for a legal
            configuration, but it might be illegal because
            it contains gaps or its edge is not all blank. *)
         fun checklegal (x0, y0, x1, y1) =
           let in
             (* interior *)
             util_for y0 y1
             (fn y =>
              util_for x0 x1
              (fn x =>
               if test (x, y)
               then ()
               (* Could constrain the illegal area to the first interior gap, but that
                  might not contain the start point. *)
               else raise Return (IllegalConfig { x = x0, y = y0, 
                                                  w = x1 - x0 + 1, h = y1 - y0 + 1 })
               ));

             (* left and right edge *)
             util_for y0 y1
             (fn y =>
              if test (x0 - 1, y)
              then if test (x1 + 1, y)
                   then ()
                   else raise Return (IllegalConfig { x = x0, y = y0, 
                                                      w = x1 - x0 + 2, h = y1 - y0 + 1 })
              else raise Return (IllegalConfig { x = x0 - 1, y = y0, 
                                                 w = x1 - x0 + 2, h = y1 - y0 + 1 }));

             (* top and bottom edges *)
             util_for x0 x1
             (fn x =>
              if test (x, y0 - 1)
              then if test (x, y1 + 1)
                   then ()
                   else raise Return (IllegalConfig { x = x0, y = y0,
                                                      w = x1 - x0 + 2, h = y1 - y0 + 2 })
              else raise Return (IllegalConfig { x = x0, y = y0 - 1,
                                                 w = x1 - x0 + 1, h = y1 - y0 + 2 }))
           end
     in
         checklegal (x, y, xx, yy);
         Battleship { x = x, y = y, w = xx - x + 1, h = yy - y + 1 }
     end handle Return s => s

end