
structure Clip :> CLIP =
struct

   (* in pixels *)
   type config = { width : int,
                   height : int }

   (* Note: the snail movement code also assumes this shape. *)
   val horiz = { width = 3 * Tile.TILEW,
                 height = 2 * Tile.TILEH }

   val vert = { width = 2 * Tile.TILEW,
                height = 3 * Tile.TILEH }

   val bullet = { width = 23, height = 10 }
   val star = { width = 32, height = 32 }
   val spider = { width = 25, height = 21 }
   val rock = { width = 32, height = 32 }
   val unit = { width = 1, height = 1 }

   fun center { width, height } = (width div 2, height div 2)
   fun clip_width { width, height } = width

    datatype clipobj =
        ClipRoom 
      | ClipRock of Types.rock ref

   (* PERF slow! should precompute this! (how?) *)
   (* could use a mipmap for the mask region? *)
   (* not possible to occupy this space if any
      pixel within the configuration rectangle 
      is solid. *)
   exception EarlyClip of clipobj
   fun clipped_room room { width, height } (px, py) =
       (Util.for 0 (width - 1)
        (fn xx =>
         Util.for 0 (height - 1)
         (fn yy =>
          let
              val x = xx + px
              val y = yy + py
              (* x,y are world-absolute coordinates.
                 we need the tile that's at those
                 coordinates: *)
              val tx = x div Tile.TILEW
              val ty = y div Tile.TILEH
              (* and the offset within that tile: *)
              val ox = x mod Tile.TILEW
              val oy = y mod Tile.TILEH
          in
              if Tile.clipmask (Room.maskat (room, tx, ty)) (ox, oy)
              then raise (EarlyClip ClipRoom)
              else ()
          end));
        
        (* if everything is clear with no exn, then ok *)
        []
        ) handle EarlyClip x => [x]

   fun clipped_rocks rocks { width, height } (px, py) =
       let
         fun clipped_rock (hitrock as (ref (Types.Rock { x, y, dx, dy }))) =
           let val result = 
             (Util.for 0 (width - 1)
              (fn xx =>
               Util.for 0 (height - 1)
               (fn yy =>
                let
                    val objx = xx + px
                    val objy = yy + py
                    val { width = rockw, height = rockh } = rock
                in
                    if objx > x andalso objy > y andalso
                       objx < x + rockw andalso objy < y + rockh
                    then raise EarlyClip (ClipRock hitrock)
                    else ()
                end));
              NONE
              ) handle EarlyClip x => SOME x
              (*
             val _ = print ("Clipping against a rock: we are (" ^ (Int.toString
               width) ^ ", " ^ (Int.toString height) ^ "), rock is (" ^ (Int.toString x)
               ^ ", " ^ (Int.toString y) ^ "), result is " ^ (if isSome result
               then "SOME" else "NONE") ^ "\n")
               *)
           in
             result
           end
       in
         List.mapPartial clipped_rock rocks
       end

   fun cliplist room rocks clip pos =
     let 
       fun printclip ClipRoom = print "ROOM "
         | printclip (ClipRock (ref (Types.Rock { x, y, dx, dy }))) = print "ROCK "
       val clippings = (clipped_room room clip pos) @ (clipped_rocks rocks clip pos)
       (*
       val _ = print "Clippings: "
       val _ = map printclip clippings
       val _ = print "\n"
       *)
     in
       clippings
     end

   fun clipped room rocks clip pos = not (List.null (cliplist room rocks clip pos))
end

