structure Linereal :> LINEREAL =
struct
  structure W = Word32
  structure I = Int
  val w2i = W.toIntX
  val i2w = W.fromInt

  type coord = int * int
  type t = I.int * I.int * I.int

  fun build ((x0,y0), (x1,y1), (dx,dy), (stepx,stepy), post) =
    let val frac0 = dy - Int.quot (dx, 2)
        fun step (x0,y0,frac) =
          if x0 = x1
          then NONE
          else
              let val (y0,frac) = if frac >= 0 then (y0 + stepy, frac - dx) else (y0, frac)
                  val x0 = x0 + stepx
                  val frac = frac + dy
              in SOME ((x0,y0,frac), post (x0,y0)) end
    in ({step=step, seed=(x0,y0,frac0)}, post (x0,y0)) end

  fun linereal p0 p1 =
    let
        val d = Pair.map2 I.- p1 p0
        fun abs c = if c < 0 then (~c,~1) else (c,1)
        val ((dx',stepx),(dy',stepy)) = Pair.map abs d
        val step = (stepx,stepy)
        val d'' as (dx'',dy'') = Pair.map (fn n => n*2) (dx',dy')

        val swap = Pair.swap
        val cvt = Pair.map (fn x => x)
        val seed =
            if dx'' > dy''
            then (p0, p1, d'', step, cvt)
            else (swap p0, swap p1, swap d'', swap step, swap)
    in build seed end

  fun all pred p0 p1 =
    let val ({step, seed},v) = linereal p0 p1
        fun loop seed =
            case step seed
             of NONE => true
              | SOME (seed', v) => pred v andalso loop seed'
    in pred v andalso loop seed end

end
